import asyncio
import functools

import aioredis
from aioredis.pubsub import Receiver

from aiopubsub.base import BasePubsub


def init_conn(func):
    @functools.wraps(func)
    async def wrapper(self, *args, _conn=None, **kwargs):
        if _conn is None:
            pool = await self._get_pool()
            conn_context = await pool
            with conn_context as _conn:
                _conn = aioredis.Redis(_conn)
                return await func(self, *args, _conn=_conn, **kwargs)
        return await func(self, *args, _conn=_conn, **kwargs)

    return wrapper


class RedisBackend:
    def __init__(self, host="127.0.0.1", port=6379, db=0, password=None,
                 loop=None, socket_connect_timeout=None,
                 pool_min_size=5, pool_max_size=20,
                 **kwargs):
        super().__init__(**kwargs)
        self.host = host
        self.port = int(port)
        self.db=db
        self._loop = loop
        create_connection_timeout = (float(socket_connect_timeout) if socket_connect_timeout else None)
        self.kwargs = {"db": int(db), "password": password,
                       "create_connection_timeout": create_connection_timeout,
                       "encoding": "utf-8",
                       "minsize": pool_min_size,
                       "maxsize": pool_max_size,
                       }

        self.__pool_lock = None
        self._pool: aioredis.ConnectionsPool = None

    @property
    def _pool_lock(self):
        if self.__pool_lock is None:
            self.__pool_lock = asyncio.Lock()
        return self.__pool_lock

    async def _acquire_sub(self):
        await self._get_pool()
        conn = await self._pool.acquire()
        conn = aioredis.Redis(conn)
        return conn

    async def _release_sub(self, _conn: aioredis.Redis):
        self._pool.release(_conn.connection)
        return

    async def _acquire_pub(self):
        await self._get_pool()
        conn = await self._pool.acquire()
        conn = aioredis.Redis(conn)
        return conn

    async def _release_pub(self, _conn: aioredis.Redis):
        self._pool.release(_conn.connection)
        return

    @init_conn
    async def _unsubscribe(self, channel, *channels, _conn: aioredis.Redis = None):
        return await _conn.unsubscribe(channel, *channels)

    @init_conn
    async def _subscribe(self, channel, *channels, _conn: aioredis.Redis = None):
        mpsc = Receiver()
        ns_channels = [mpsc.channel(c) for c in (channel, *channels)]
        await _conn.subscribe(*ns_channels)
        return mpsc

    @init_conn
    async def _psubscribe(self, pattern, *patterns, _conn: aioredis.Redis = None):
        mpsc = Receiver()
        ns_patterns = [mpsc.pattern(p) for p in (pattern, *patterns)]
        await _conn.psubscribe(*ns_patterns)
        return mpsc

    @init_conn
    async def _punsubscribe(self, pattern, *patterns, _conn: aioredis.Redis = None):
        return await _conn.punsubscribe(pattern, *patterns)

    async def _listen(self, _conn: Receiver = None):
        """Listen for messages on channels this client has been subscribed to"""
        async for channel, msg in _conn.iter():
            if channel.is_pattern:
                channel_name, data = msg
                channel_name = channel_name.decode()
                data = data.decode()
                pattern = channel.name.decode()
                type_ = "pmessage"
            else:
                pattern = None
                type_ = "message"
                data = msg.decode()
                channel_name = channel.name.decode()
            yield {"type": type_, "pattern": pattern, "channel": channel_name, "data": data}

    @init_conn
    async def _publish(self, channel, message, _conn: aioredis.Redis = None):
        return await _conn.publish(channel, message)

    async def _close(self, *args, **kwargs):
        if self._pool is not None:
            await self._pool.clear()

    async def _get_pool(self):
        async with self._pool_lock:
            if self._pool is None:
                self._pool = await aioredis.create_pool((self.host, self.port), **self.kwargs)
            return self._pool


class RedisPubsub(RedisBackend, BasePubsub):
    NAME = "redis"

    def __repr__(self):
        return f"RedisPubsub ({self.host}:{self.port}/{self.db})"



import logging
from typing import Dict, Type

from .base import BasePubsub

__version__ = "1.0.2"


def int_or_str(value):
    try:
        return int(value)
    except ValueError:
        return value


VERSION = tuple(map(int_or_str, __version__.split(".")))
LOG = logging.getLogger(__name__)

PUBSUB_CACHES: Dict[str, Type[BasePubsub]] = {}

try:
    import aioredis
except ImportError:
    LOG.info("aioredis not installed, RedisCache unavailable")
else:
    from .backend import RedisPubsub

    PUBSUB_CACHES[RedisPubsub.NAME] = RedisPubsub
    del aioredis


class Pubsub:
    REDIS = PUBSUB_CACHES[RedisPubsub.NAME]

    def __new__(cls, pubsub_class=REDIS, **kwargs):
        if not issubclass(pubsub_class, BasePubsub):
            raise TypeError("Invalid cache type, you can only use {}".format(list(PUBSUB_CACHES.keys()))
                            )
        instance = pubsub_class.__new__(pubsub_class, **kwargs)
        instance.__init__(**kwargs)
        return instance


__all__ = [
    "Pubsub"
]



import asyncio
import functools

import aioredis

from aiopubsub.base import BasePubsub


def init_pub(func):
    @functools.wraps(func)
    async def wrapper(self, *args, _conn=None, **kwargs):
        if _conn is None:
            redis = await self._get_redis()
            async with redis.client() as _conn:
                return await func(self, *args, _conn=_conn, **kwargs)
        return await func(self, *args, _conn=_conn, **kwargs)

    return wrapper


def init_sub(func):
    @functools.wraps(func)
    async def wrapper(self, *args, _conn=None, **kwargs):
        if _conn is None:
            redis = await self._get_redis()
            async with redis.pubsub() as _conn:
                return await func(self, *args, _conn=_conn, **kwargs)
        return await func(self, *args, _conn=_conn, **kwargs)

    return wrapper


class RedisBackend:
    def __init__(self, host="127.0.0.1", port=6379, db=0, password=None,
                 loop=None, socket_connect_timeout=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.host = host
        self.port = int(port)
        self.db = db
        self._loop = loop
        socket_connect_timeout = (float(socket_connect_timeout) if socket_connect_timeout else None)
        self.kwargs = {"db": int(db), "password": password, "decode_responses": True,
                       "socket_connect_timeout": socket_connect_timeout}

        self.__redis_lock = None

        self._redis: aioredis.Redis = None

    @property
    def _redis_lock(self):
        if self.__redis_lock is None:
            self.__redis_lock = asyncio.Lock()
        return self.__redis_lock

    async def _acquire_sub(self):
        await self._get_redis()
        return self._redis.pubsub()

    async def _release_sub(self, _conn: aioredis.client.PubSub):
        return await _conn.reset()

    async def _acquire_pub(self):
        await self._get_redis()
        return await self._redis.client().initialize()

    async def _release_pub(self, _conn: aioredis.Redis):
        return await _conn.close()

    @init_sub
    async def _unsubscribe(self, channel, *channels, _conn: aioredis.client.PubSub = None):
        return await _conn.unsubscribe(channel, *channels)

    @init_sub
    async def _subscribe(self, channel, *channels, _conn: aioredis.client.PubSub = None):
        await _conn.subscribe(channel, *channels)
        return _conn

    @init_sub
    async def _psubscribe(self, pattern, *patterns, _conn: aioredis.client.PubSub = None):
        await _conn.psubscribe(pattern, *patterns)
        return _conn

    @init_sub
    async def _punsubscribe(self, pattern, *patterns, _conn: aioredis.client.PubSub = None):
        return await _conn.punsubscribe(pattern, *patterns)

    async def _listen(self, _conn: aioredis.client.PubSub = None):
        """Listen for messages on channels this client has been subscribed to"""
        async for k in _conn.listen():
            if not k["type"] in ["pmessage", "message"]:
                continue
            yield k

    @init_pub
    async def _publish(self, channel, message, _conn: aioredis.client.Redis = None):
        return await _conn.publish(channel, message)

    async def _close(self, *args, **kwargs):
        if self._redis is not None:
            await self._redis.close()

    async def _get_redis(self):
        async with self._redis_lock:
            if self._redis is None:
                url = f"redis://{self.host}:{self.port}"
                self._redis = aioredis.from_url(url, **self.kwargs)
            return self._redis


class RedisPubsub(RedisBackend, BasePubsub):
    NAME = "redis"

    def __repr__(self):
        return f"RedisPubsub ({self.host}:{self.port}/{self.db})"



import asyncio
import functools
import logging

from .serializer import JsonSerializer

LOG = logging.getLogger(__name__)


def process_timeout(func):
    not_set = "NOT_SET"

    @functools.wraps(func)
    async def _timeout(self, *args, timeout=not_set, **kwargs):
        timeout = self.timeout if timeout == not_set else timeout
        if timeout == 0 or timeout is None:
            return await func(self, *args, **kwargs)
        return await asyncio.wait_for(func(self, *args, **kwargs), timeout)

    return _timeout


class BasePubsub(object):
    NAME: str

    def __init__(self, serializer=None, namespace=None, key_builder=None, timeout=500):
        self.namespace = namespace
        self.timeout = timeout
        self.build_key = key_builder or self._build_key
        self._serializer = serializer or JsonSerializer()

    @property
    def serializer(self):
        return self._serializer

    def _build_key(self, key, namespace=None):
        if namespace is not None:
            return f"{namespace}:{key}"
        if self.namespace is not None:
            return f"{self.namespace}:{key}"
        return key

    @process_timeout
    async def unsubscribe(self, channel, *channels, namespace=None, _conn=None):
        ns_channels = [self.build_key(c, namespace=namespace) for c in (channel, *channels)]
        return await self._unsubscribe(*ns_channels, _conn=_conn)

    @process_timeout
    async def publish(self, channel, data, dumps_fn=None, namespace=None, _conn=None):
        dumps = dumps_fn or self._serializer.dumps
        ns_channel = self.build_key(channel, namespace=namespace)
        message = dumps(data)
        ret = await self._publish(ns_channel, message, _conn=_conn)
        if ret == 0:
            LOG.warning(f"No client subscribe channel[{ns_channel}], message is dropped: {message}")
        return ret

    @process_timeout
    async def subscribe(self, channel, *channels, _conn=None, namespace=None):
        ns_channels = [self.build_key(c, namespace=namespace) for c in (channel, *channels)]
        return await self._subscribe(*ns_channels, _conn=_conn)

    @process_timeout
    async def psubscribe(self, pattern, *patterns, _conn=None, namespace=None):
        ns_channels = [self.build_key(c, namespace=namespace) for c in (pattern, *patterns)]
        return await self._psubscribe(*ns_channels, _conn=_conn)

    @process_timeout
    async def punsubscribe(self, pattern, *patterns, namespace=None, _conn=None):
        ns_channels = [self.build_key(c, namespace=namespace) for c in (pattern, *patterns)]
        return await self._punsubscribe(*ns_channels, _conn=_conn)

    async def listen(self, loads_fn=None, _conn=None):
        loads = loads_fn or self._serializer.loads
        async for k in self._listen(_conn):
            k["data"] = loads(k["data"])
            yield k

    @process_timeout
    async def unsubscribe(self, channel, *channels, _conn=None, namespace=None):
        ns_channels = [self.build_key(c, namespace=namespace) for c in (channel, *channels)]
        return await self._unsubscribe(*ns_channels, _conn=_conn)

    @process_timeout
    async def close(self, *args, _conn=None, **kwargs):
        return await self._close(*args, _conn=_conn, **kwargs)

    def get_pub(self):
        return _Pub(self)

    def get_sub(self):
        return _Sub(self)

    async def acquire_sub(self):
        return await self._acquire_sub()

    async def release_sub(self, _conn):
        return await self._release_sub(_conn)

    async def acquire_pub(self):
        return await self._acquire_pub()

    async def release_pub(self, _conn):
        return await self._release_pub(_conn)


class _Pub(object):
    def __init__(self, _pubsub):
        self._pubsub = _pubsub
        self.conn = None

    async def __aenter__(self):
        self.conn = await self._pubsub.acquire_pub()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self._pubsub.release_pub(self.conn)

    def __getattr__(self, name):
        return self._pubsub.__getattribute__(name)


class _Sub(object):
    def __init__(self, _pubsub):
        self._pubsub = _pubsub
        self.conn = None

    async def __aenter__(self):
        self.conn = await self._pubsub.acquire_sub()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self._pubsub.release_sub(self.conn)

    def __getattr__(self, name):
        return self._pubsub.__getattribute__(name)



import logging
from typing import Any, Optional

LOG = logging.getLogger(__name__)

try:
    import ujson as json
except ImportError:
    LOG.debug("ujson module not found, using json")
    import json

_NOT_SET = object()


class BaseSerializer:
    DEFAULT_ENCODING: Optional[str] = "utf-8"

    def __init__(self, *args, encoding=_NOT_SET, **kwargs):
        self.encoding = self.DEFAULT_ENCODING if encoding is _NOT_SET else encoding
        super().__init__(*args, **kwargs)

    def dumps(self, value: Any) -> str:
        raise NotImplementedError("dumps method must be implemented")

    def loads(self, value: str) -> Any:
        raise NotImplementedError("loads method must be implemented")


class JsonSerializer(BaseSerializer):
    def dumps(self, value):
        return json.dumps(value)

    def loads(self, value):
        if value is None:
            return None
        return json.loads(value)



import aioredis
import re

__all__ = ["RedisPubsub"]


def get_redis_version():
    version1_reg = re.compile(r"^1.*")
    version2_reg = re.compile(r"^[2-9].*")
    redis_version = aioredis.__version__
    if version1_reg.match(redis_version):
        return 1
    elif version2_reg.match(redis_version):
        return 2
    return 0


_version = get_redis_version()
if _version == 1:
    from .redis import RedisPubsub
elif _version == 2:
    from .redis2 import RedisPubsub
else:
    raise ImportError(f"unsupported version aioredis: {aioredis.__version__}")
